[].pop('not an int')
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__pop_type_error.py", line 1, in <module>
    [].pop('not an int')
    ~~~~~~~~~~~~~~~~~~~~
TypeError: 'str' object cannot be interpreted as an integer
"""
