x = True
if x:
    continue
"""
TRACEBACK:
Traceback (most recent call last):
  File "loop__continue_in_if_error.py", line 3
    continue
    ~~~~~~~~
SyntaxError: 'continue' not properly in loop
"""
