a, b = (1, 2, 3, 4, 5)
"""
TRACEBACK:
Traceback (most recent call last):
  File "unpack__too_many.py", line 1, in <module>
    a, b = (1, 2, 3, 4, 5)
    ~~~~
ValueError: too many values to unpack (expected 2, got 5)
"""
