lst = [1, 2, 3]
lst[10] = 'value'
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__setitem_index_error.py", line 2, in <module>
    lst[10] = 'value'
    ~~~~~~~
IndexError: list assignment index out of range
"""
