(1, 2, 3).index(4)
"""
TRACEBACK:
Traceback (most recent call last):
  File "tuple__index_not_found.py", line 1, in <module>
    (1, 2, 3).index(4)
    ~~~~~~~~~~~~~~~~~~
ValueError: tuple.index(x): x not in tuple
"""
