'hello'.partition('')
"""
TRACEBACK:
Traceback (most recent call last):
  File "str__partition_empty.py", line 1, in <module>
    'hello'.partition('')
    ~~~~~~~~~~~~~~~~~~~~~
ValueError: empty separator
"""
