float("a\nb")
"""
TRACEBACK:
Traceback (most recent call last):
  File "type__float_repr_newline.py", line 1, in <module>
    float("a\nb")
    ~~~~~~~~~~~~~
ValueError: could not convert string to float: 'a\nb'
"""
