[1, 2, 3].pop(10)
"""
TRACEBACK:
Traceback (most recent call last):
  File "list__pop_out_of_range.py", line 1, in <module>
    [1, 2, 3].pop(10)
    ~~~~~~~~~~~~~~~~~
IndexError: pop index out of range
"""
