slice(1, 2, 3, 4)
"""
TRACEBACK:
Traceback (most recent call last):
  File "slice__too_many_args.py", line 1, in <module>
    slice(1, 2, 3, 4)
    ~~~~~~~~~~~~~~~~~
TypeError: slice expected at most 3 arguments, got 4
"""
